"""Implementation of lpmn client."""
import argparse
import ast
import time

import requests
from requests.exceptions import HTTPError

from lpmn_client_biz import Connection, IOType, Task, delete, download, upload

requests.packages.urllib3.disable_warnings(
    requests.packages.urllib3.exceptions.InsecureRequestWarning)


def arg_as_list(s):
    """Parse string list to list.

    :param s: String containing list
    :type s: str
    :return: list
    """
    v = ast.literal_eval(s)
    if type(v) is not list:
        raise argparse.ArgumentTypeError(
            f"Argument {s} is not a list")  # pragma: no cover
    return v


def get_args():
    """Parse arguments."""
    parser = argparse.ArgumentParser(
        description="LPMN CLI tool",
    )
    parser.add_argument(
        "lpmn",
        help="LPMN json query",
        type=arg_as_list
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="If progress should be printed.",
        action="store_true"
    )
    parser.add_argument(
        "input",
        help="Path to a input file on local machine.",
    )
    parser.add_argument(
        "-it",
        "--input-type",
        help="Type of input",
        type=IOType.argparse,
        choices=list(IOType),
        default=IOType.TEXT
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Path to a output directory on local machine",
        default=''
    )
    parser.add_argument(
        "-ot",
        "--output-type",
        help="Type of output",
        type=IOType.argparse,
        choices=list(IOType),
        default=IOType.FILE
    )
    parser.add_argument(
        "-f",
        "--filename",
        help="Downloaded filename",
        default=''
    )
    parser.add_argument(
        "-t",
        "--timeout",
        help="Task timeout",
        default=0,
        type=int
    )
    parser.add_argument(
        "-u",
        "--username",
        help="E-science username",
        default=''
    )
    parser.add_argument(
        "-p",
        "--password",
        help="E-science user password",
        default=''
    )
    parser.add_argument(
        "--url",
        help="Base api url",
        default=''
    )
    parser.add_argument(
        "-c",
        "--config",
        help="Path to config file",
        default=''
    )
    return parser.parse_args()


def main():
    """Runs the program."""
    task_args = get_args()
    try:
        connection = Connection(task_args.username, task_args.password,
                                task_args.url, task_args.config)
    except Exception as e:  # pragma: no cover
        print(f"{str(e)}\n"
              f"Maybe you forgot to provide "
              f"username and password or config file")  # pragma: no cover
    else:
        task = Task(task_args.lpmn, connection)

        task_input, result = None, None
        if task_args.input_type == IOType.FILE:
            print(f"Uploading {task_args.input} to server")
            try:
                task_input = upload(connection, task_args.input)
            except HTTPError as e:  # pragma: no cover
                print(f"Uploading failed. {str(e)} "
                      f"Could not upload "
                      f"{task_args.input} to server")  # pragma: no cover
            except FileNotFoundError:  # pragma: no cover
                print(f"No such file or directory: "
                      f"{task_args.input}")  # pragma: no cover
        else:
            task_input = task_args.input

        if task_input:
            start_time = time.time()
            if task_args.verbose:
                print("Running task")
            try:
                output_file_id = task.run(task_input, task_args.input_type,
                                          verbose=task_args.verbose,
                                          timeout=task_args.timeout)
                if task_args.input_type == IOType.FILE:
                    delete(connection, task_input)
                if task_args.verbose:
                    print(f"Task finished in "
                          f"{round(time.time() - start_time, 2)} seconds")
            except HTTPError as e:  # pragma: no cover
                print(f"Task execution failed for parameters "
                      f"LPMN {task_args.lpmn}, "
                      f"Input {task_input} of type {task_args.input_type}"
                      f"\nDetails {str(e)}")  # pragma: no cover
            else:
                if task_args.output_type == IOType.FILE_ID:
                    result = output_file_id
                else:
                    if task_args.verbose:
                        print("Downloading result")
                    try:
                        result = download(connection,
                                          file_id=output_file_id,
                                          output_type=task_args.output_type,
                                          output_path=task_args.output,
                                          filename=task_args.filename)
                    except HTTPError:  # pragma: no cover
                        print(f"Downloading {output_file_id} \
                         from server failed")  # pragma: no cover
                if result:
                    return result
