"""Task implementation."""
import _thread
import configparser
import os
import threading
import time
from contextlib import contextmanager
from typing import Optional

import requests
from tqdm import tqdm

from lpmn_client_biz import Connection, IOType

PROGRESS_SLEEP_PERIOD = 0.5

DONE = 'DONE'
ERROR = 'ERROR'
PROCESSING = 'PROCESSING'
NOT_RUNNING_STATUSES = ['QUEUE', 'NONEXISTING']
CANCEL = 'CANCEL'

UPLOADED = 'Uploaded'

FileID = str
SUPPORTED_HEADER_ACTIONS = ['run_sent']


@contextmanager
def _timeout_manager(timeout, task, msg=''):
    timer = threading.Timer(timeout, lambda: _thread.interrupt_main())
    if timeout > 0:
        timer.start()
    try:
        yield
    except KeyboardInterrupt:
        task.cancel()
        raise TimeoutError(f"Operation {msg} timeout")
    finally:
        timer.cancel()


class Task:
    """Implements Task class."""

    def __init__(self, lpmn: list, connection: Connection) -> None:
        """Task init file.

        :param lpmn: json style list with lpmn commands
        :type lpmn: list
        """
        self.status = None
        self.lpmn = lpmn
        self.task_id = None
        self.__connection = connection

    def __get_task_info(self):
        endpoint = Connection.API_ENDPOINTS['tasks']['progress']
        if self.task_id:
            r = requests.get(
                url=f"{self.__connection.base_url}{endpoint}{self.task_id}",
                headers=self.__connection.headers(),
                verify=False
            )
            if r.status_code >= 400 and "detail" in r.json():
                raise requests.HTTPError(f'{r.status_code} '
                                         f'Client Error: {r.json()["detail"]}')
            else:
                r.raise_for_status()
            return r.json()
        else:
            raise Exception("Task not started yet")

    def get_progress(self) -> float:
        """Returns progress of the running task.

        :return: task progress value [0, 1]
        :rtype: float
        """
        data = self.__get_task_info()
        if data:
            status = data["status"]
            if status == ERROR:
                raise Exception(f"ERROR ({data})")  # pragma: no cover
            elif status in NOT_RUNNING_STATUSES:
                progress = 0      # pragma: no cover
            elif status == PROCESSING:
                progress = data["progress"]  # pragma: no cover
            elif status == CANCEL:
                raise Exception("Task has been canceled")  # pragma: no cover
            else:
                progress = 1
            return progress

    def _wait(self, verbose: bool = False):
        """Wait for task to be finished.

        :param timeout: task waiting timeout in seconds
        :type timeout: float
        :param verbose: flag if task should return progress
        :type verbose: bool
        """
        status = self.__get_task_info()['status']
        pbar = tqdm(total=100) if verbose else None

        while status != DONE:
            time.sleep(0.5)
            info = self.__get_task_info()
            progress, status = info['progress'], info['status']
            if pbar:
                pbar.update(progress * 100 - pbar.n)

        if pbar:
            pbar.update(100 - pbar.n)
        self.status = DONE

    def cancel(self):
        """Cancel running task."""
        endpoint = Connection.API_ENDPOINTS['tasks']['cancel']
        if self.task_id:
            r = requests.get(
                url=f"{self.__connection.base_url}{endpoint}{self.task_id}",
                headers=self.__connection.headers(),
                verify=False
            )
            if r.status_code >= 400 and "detail" in r.json():
                raise requests.HTTPError(f'{r.status_code} '
                                         f'Client Error: {r.json()["detail"]}')
            else:
                r.raise_for_status()
            self.status = CANCEL
        else:
            raise Exception("Task not started yet")

    def get_output(self) -> Optional[str]:
        """Returns task output id, raises exception if task is not completed.

        :return: output file_id if task is completed
        :rtype: str, optional
        """
        info = self.__get_task_info()
        if info['status'] == DONE:
            return self.__get_task_info()['output']
        else:
            raise Exception("Task is not completed yet")  # pragma: no cover

    def run_sent(self, texts: list) -> Optional[list]:
        """Start sentence task execution.

        :param texts: list of input sentences
        :type texts: list
        :return: output list of vectors
        :rtype: list, optional
        """
        endpoint = Connection.API_ENDPOINTS['tasks']['sent']
        r = requests.post(
            url=f"{self.__connection.base_url}{endpoint}",
            headers=self.__connection.headers(),
            json={
                "application": 'sent_client',
                "task": self.lpmn[0],
                "input": texts,
            },
            verify=False
        )
        if r.status_code == 404 and "detail" in r.json():
            raise requests.HTTPError(f'404 Client Error: {r.json()["detail"]}')
        else:
            r.raise_for_status()
        return r.json()

    def run(self,
            input: str,
            input_type: IOType,
            blocking: bool = True,
            timeout: float = 0,
            verbose: bool = False
            ) -> Optional[str]:
        """Start task execution.

        :param input: input data file_id or text
        :type input: str
        :param input_type: type of input text or file
        :type input_type: InputType
        :param blocking: if it is blocking method will wait for
         task to finish and return output file_id
        :type blocking: bool
        :param timeout: timeout for blocking task
        :type timeout: float
        :param verbose: if blocking task progress should be returned
        :type verbose: bool
        :return: output file_id or None if not blocking task
        :rtype: str, optional
        """
        input_type_value = input_type.value if \
            input_type != IOType.FILE_ID else IOType.FILE.value

        endpoint = Connection.API_ENDPOINTS['tasks']['start']
        r = requests.post(
            url=f"{self.__connection.base_url}{endpoint}",
            headers=self.__connection.headers(),
            json={
                "application": "lpmn_client",
                "task": self.lpmn,
                "input": input,
                "input_type": input_type_value
            },
            verify=False
        )
        if r.status_code >= 400 and "detail" in r.json():
            raise requests.HTTPError(f'{r.status_code} '
                                     f'Client Error: {r.json()["detail"]}')
        else:
            r.raise_for_status()
        data = r.json()
        self.task_id = data['task_id']

        if blocking:
            with _timeout_manager(timeout, self, self.lpmn):
                self._wait(verbose=verbose)
            return self.get_output()


def run_with_header(config_path, x_token, lpmn, action, *action_args):
    """Runs given action with arguments, with x-token header.

    :param config_path: path to config.ini file
    :type config_path: str
    :param x_token: api authorization token
    :type x_token: str
    :param lpmn: json style list with lpmn commands
    :type lpmn: list(str)
    :param action: Task class method name
    :type action: str
    :param action_args: arguments for Task method
    :return: action result
    """
    base_url, yaml_config = '', ''
    if os.path.exists(config_path) and os.path.isfile(config_path):
        config = configparser.ConfigParser()
        config.read(config_path)
        base_url = config.get('tool', 'base_url', fallback='')
        yaml_config = config.get('tool', 'yaml_config', fallback='')

    x_header = True if base_url else False
    if not x_token:
        x_token = Connection(config_file=yaml_config).token

    conn = Connection(baseurl=base_url,
                      x_token=x_token,
                      x_token_header=x_header)
    task = Task(lpmn, conn)
    if action in SUPPORTED_HEADER_ACTIONS:
        return getattr(task, action)(*action_args)
    else:
        raise Exception("Not supported yet")
